import os
import re
from pathlib import Path
from typing import Dict

from kaggle import KaggleApi
from kaggle.models.kaggle_models_extended import KernelPushResponse

from .. import kernel_proc
from ..resource import get_dataset_slug
from ..builders import get_builder
from ..exception import MetaDataNotFound
from ..repo import Repo
from .kkt_command import kkt_command


def create_kernel_body(meta_data: Dict, pkg_dataset, env_variables: Dict) -> str:
    enable_internet = meta_data.get("enable_internet", False)
    enable_constraint = meta_data.get("enable_constraint", False)
    kernel_type = meta_data.get("kernel_type", "script")
    code_file_path = Path(meta_data.get("code_file", "main.py"))
    if not code_file_path.is_absolute():
        code_file_path = Path.cwd() / code_file_path

    kernel_builder = get_builder(kernel_type)
    with code_file_path.open() as fp:
        return kernel_builder(
            fp, pkg_dataset, env_variables, enable_internet, enable_constraint
        )


def push_impl(
    api: KaggleApi, meta_data: Dict, env_variables: Dict
) -> KernelPushResponse:

    dataset_slug = get_dataset_slug(api, meta_data)
    dataset_name = dataset_slug.split("/")[-1]

    kernel_body = create_kernel_body(meta_data, dataset_name, env_variables)
    kernel_push_param = kernel_proc.KernelPushParams.of(api, meta_data)
    kernel_push_param.dataset_data_sources.append(dataset_slug)
    return kernel_proc.push(api, kernel_push_param, kernel_body)


def get_env_variables(env_variables: Dict) -> Dict:
    result = {**env_variables}
    for k, v in os.environ.items():
        m = re.match(r"^KKT_(.+)$", k)
        if m:
            update_key = m.groups()[0]
            result[update_key] = v
    return result


@kkt_command()
def push(api: KaggleApi, kkt: Dict, pyproject_path: Path, **kwargs: Dict) -> None:
    repo = Repo(pyproject_path.parent)
    enable_git_tag: bool = kkt.get("enable_git_tag", False)
    if enable_git_tag:
        repo.validate()

    if "meta_data" not in kkt:
        raise MetaDataNotFound()

    meta_data = kkt["meta_data"].value
    env_variables = get_env_variables(kkt.get("environment_variables", {}))

    result = push_impl(api, meta_data, env_variables)
    kernel_proc.print_response(result)

    if enable_git_tag and result.versionNumber:
        repo.attach_version_tag(result.versionNumber, meta_data, env_variables)



from typing import List
from pathlib import Path

from dataclasses import dataclass
import requests


@dataclass
class PackageLocation:
    url: str
    name: str


def fetch_packages(
    locations: List[PackageLocation], save_dir: Path, quiet: bool = False
) -> List[Path]:
    outfiles = []
    for loc in locations:
        response = requests.get(loc.url)
        outfile = save_dir / loc.name
        with outfile.open("wb") as out:
            out.write(response.content)
        outfiles.append(outfile)
        if not quiet:
            print("Output file downloaded to %s" % str(outfile))
    return outfiles



import json
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List

import click
from kaggle import KaggleApi
from kaggle.models.kaggle_models_extended import KernelPushResponse

from .. import kernel_proc
from ..builders.packaging_system import get_dependencies
from ..exception import InstallKernelError, MetaDataNotFound
from ..resource import get_username, get_dataset_slug
from .kkt_command import kkt_command
from ..fetch import PackageLocation, fetch_packages


def create_kernel_body(install_pkgs: List[str]) -> str:
    return f"""
import subprocess

def pip_freeze():
    args = ["pip", "freeze"]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE)
    run_result = proc.communicate()
    return run_result[0].decode("utf-8").split("\\n")

def pip_install(pkgs):
    if len(pkgs) == 0:
        return
    args = ["pip", "install", *pkgs]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE)
    return proc.communicate()[0].decode("utf-8")

def pip_download(pkgs):
    if len(pkgs) == 0:
        return ""
    args = ["pip", "download", "--no-deps", *pkgs]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE)
    return proc.communicate()[0].decode("utf-8")

freeze_before_install = pip_freeze()
print(pip_install({install_pkgs}))
freeze_after_install = pip_freeze()

diff_pkgs = set(freeze_after_install) - set(freeze_before_install)
print(pip_download(diff_pkgs))
"""


def create_kernel_push_params(
    api: KaggleApi, meta_data: Dict
) -> kernel_proc.KernelPushParams:
    slug = meta_data["slug"]
    install_kernel_slug = f"{slug}-install"
    install_kernel_meta_data = {
        **meta_data,
        "slug": install_kernel_slug,
        "kernel_type": "script",
        "is_private": True,
        "enable_gpu": False,
        "enable_internet": True,
        "dataset_sources": [],
        "competition_sources": [],
        "kernel_sources": [],
        "keywords": [],
    }
    return kernel_proc.KernelPushParams.of(api, install_kernel_meta_data)


def get_owner_slug_from(response: KernelPushResponse):
    return response.ref.split("/")[1]


def get_kernel_slug_from(response: KernelPushResponse):
    return response.ref.split("/")[2]


def get_error_messages(logs: Dict) -> List[str]:
    result = []
    for log in logs:
        stream_name = log.get("stream_name", "stderr")
        data = log.get("data", "")
        if stream_name == "stderr" and not (
            data.startswith("[NbConvertApp]") or data.startswith("WARNING:")
        ):
            result.append(data)
    return result


def _get_package_locations(list_response: Dict[str, Any]) -> List[PackageLocation]:
    return [
        PackageLocation(item["url"], item["fileName"])
        for item in list_response["files"]
    ]


def wait_for_install_kernel_completion(
    api: KaggleApi, kernel_slug: str, quiet: bool = False
) -> Dict[str, Any]:
    while True:
        owner_slug = get_username(api)
        response = api.process_response(
            api.kernel_output_with_http_info(owner_slug, kernel_slug)
        )

        if response["log"] != "":
            logs = json.loads(response["log"])
            err_messages = get_error_messages(logs)
            if 0 < len(err_messages):
                raise InstallKernelError(err_messages)
            if 0 < len(logs):
                return response

        if not quiet:
            click.echo("Wait for install kernel completion...")
        time.sleep(10)


def upload_requirement_pkgs(
    api: KaggleApi, meta_data: Dict, target_dir: Path, quiet: bool = False
):
    slug = get_dataset_slug(api, meta_data)
    _, dataset_slug = slug.split("/")[-2:]
    license_name = "CC0-1.0"
    status = api.dataset_status(slug)
    if status is None:
        return kernel_proc.create_dataset(
            api,
            dataset_slug=dataset_slug,
            license_name=license_name,
            target_dir=target_dir,
            quiet=quiet,
        )
    else:
        return kernel_proc.update_dataset(
            api, dataset_slug=dataset_slug, target_dir=target_dir, quiet=quiet,
        )


def push_install_kernel(
    api: KaggleApi, meta_data: Dict, quiet: bool = False
) -> KernelPushResponse:
    enable_constraint = meta_data.get("enable_constraint", False)

    kernel_push_params = create_kernel_push_params(api, meta_data)
    dependencies = get_dependencies(enable_constraint)
    kernel_body = create_kernel_body(dependencies)
    kernel_response = kernel_proc.push(api, kernel_push_params, kernel_body)
    if not quiet:
        kernel_proc.print_response(kernel_response)
        click.echo("Pushing install kernel successed.")

    return kernel_response


@kkt_command()
def install(
    api: KaggleApi, kkt: Dict, pyproject_path: Path, quiet: bool = False, **kwargs: Dict
) -> None:
    if "meta_data" not in kkt:
        raise MetaDataNotFound()
    meta_data = kkt["meta_data"].value
    kernel_response = push_install_kernel(api, meta_data, quiet)

    kernel_slug = get_kernel_slug_from(kernel_response)
    kernel_output = wait_for_install_kernel_completion(
        api, kernel_slug=kernel_slug, quiet=quiet
    )

    with TemporaryDirectory() as tmp_dir:
        target_dir = Path(tmp_dir)
        pkg_locations = _get_package_locations(kernel_output)
        fetch_files = fetch_packages(pkg_locations, target_dir, quiet=quiet)
        if len(fetch_files) == 0:
            click.echo("Extra required packages are nothing.")
            return

        ret = upload_requirement_pkgs(
            api, meta_data, target_dir=target_dir, quiet=quiet
        )

    kernel_proc.print_response(ret)



import json
from typing import Dict, Iterable

from .bootstrap import create_bootstrap_code


def erace_all_outputs(notebook: Dict) -> Dict:
    notebook = {**notebook}
    for cell in notebook.get("cells", []):
        if "outputs" in cell and cell["cell_type"] == "code":
            cell["outputs"] = []
    return notebook


def create_bootstrap_cell(
    pkg_encoded: str,
    pkg_dataset: str,
    env_variables: Dict,
    dependencies: Iterable[str],
    enable_internet: bool = False,
) -> Dict:
    bootstrap_code = create_bootstrap_code(
        pkg_encoded=pkg_encoded,
        pkg_dataset=pkg_dataset,
        env_variables=env_variables,
        dependencies=dependencies,
        enable_internet=enable_internet,
    )
    return {
        "cell_type": "code",
        "execution_count": None,
        "metadata": {"trusted": True, "_kg_hide-input": True, "_kg_hide-output": True},
        "outputs": [],
        "source": [bootstrap_code],
    }


def create_notebook_kernel(
    notebook_body: str,
    pkg_encoded: str,
    pkg_dataset: str,
    env_variables: Dict,
    dependencies: Iterable[str],
    enable_internet: bool = False,
) -> str:
    notebook_obj = erace_all_outputs(json.loads(notebook_body))

    bootstrap_cell = create_bootstrap_cell(
        pkg_encoded=pkg_encoded,
        pkg_dataset=pkg_dataset,
        env_variables=env_variables,
        dependencies=dependencies,
        enable_internet=enable_internet,
    )
    notebook_obj.setdefault("cells", []).insert(0, bootstrap_cell)

    notebook_kernel = json.dumps(notebook_obj)
    return notebook_kernel



from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

import click
from kaggle import KaggleApi
from kaggle.models.kaggle_models_extended import KernelPushResponse
from kaggle.models.kernel_push_request import KernelPushRequest

from ..exception import MetaDataNotFound
from ..parser import DEFAULT_KKT_CONFIG, KktParser
from .kkt_command import kkt_command
from .. import kernel_proc


def status_impl(api: KaggleApi, meta_data: Dict) -> str:
    slug = meta_data["slug"]
    result = kernel_proc.status(api, slug)

    message_elms = [f"status: {result['status']}"]
    if result["failureMessage"]:
        message_elms.append(f"message: {result['failureMessage']}")
    message = "\n".join(message_elms)
    return message


@kkt_command()
def status(api: KaggleApi, kkt: Dict, *args: List, **kwargs: Dict) -> None:
    if "meta_data" not in kkt:
        raise MetaDataNotFound()
    meta_data = kkt["meta_data"].value

    message = status_impl(api, meta_data)
    click.echo(message)



import sys
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, cast

import click
from click import Command
from kaggle import KaggleApi
from kaggle.api_client import ApiClient
from poetry.factory import Factory

from ..exception import KktSectionNotFound
from ..parser import KktParser

Wrapper = Callable[[Callable], Command]


def get_kaggle_api() -> Any:
    return KaggleApi(ApiClient())


def _wrap_click_command(f: Callable, init: bool) -> Command:
    if not init:
        f = click.option("-t", "--target", default=".", type=str)(f)
    f = click.option("-q", "--quiet", is_flag=True)(f)
    return click.command()(f)


def kkt_command(init: bool = False, cwd: Optional[Path] = None) -> Wrapper:
    def _wrapper(command: Callable) -> Command:
        @wraps(command)
        def _f(*args: List, **kwargs: Dict) -> None:
            prj_wd = Path.cwd() if cwd is None else cwd
            pyproject_path = Factory.locate(prj_wd)

            parser = KktParser(pyproject_path)

            try:
                target = cast("str", kwargs.get("target", "."))
                kkt = parser.read(key=target)
            except KktSectionNotFound:
                if not init:
                    click.echo("Kkt section is not found in pyproject.yml.", err=True)
                    sys.exit(1)
                kkt = {}

            api = get_kaggle_api()
            api.authenticate()

            command(api, kkt, pyproject_path, *args, **kwargs)

        return _wrap_click_command(_f, init)

    return _wrapper



import click

from .commands import init, install, push, status, download

cmd = click.Group()
cmd = click.version_option()(cmd)

cmd.add_command(init)
cmd.add_command(push)
cmd.add_command(status)
cmd.add_command(install)
cmd.add_command(download)


def main() -> None:
    cmd()



from typing import Callable, Dict

from .notebook_kernel import create_notebook_kernel
from .script_kernel import create_script_kernel

KERNEL_CREATEORS: Dict[str, Callable[..., str]] = {
    "script": create_script_kernel,
    "notebook": create_notebook_kernel,
}



import json
from typing import Dict, Iterable


BOOTSTRAP_TEMPLATE: str = """def __bootstrap__():
    import sys
    import base64
    import gzip
    import tarfile
    import os
    import io
    import subprocess
    from pathlib import Path
    from tempfile import TemporaryDirectory

    # install required packages
    pkg_dataset_path = Path.cwd().parent / "input" / "{pkg_dataset}"
    pkg_path_list = []
    for p in pkg_dataset_path.glob("*"):
        if p.is_dir():
            pkg_config_files = [str(p.parent) for p in p.glob("**/*") if p.name in ["pyproject.toml", "setup.py"]]
            pkg_root_dir = min(pkg_config_files, key=len)
            pkg_path_list.append(pkg_root_dir)
        else:
            pkg_path_list.append(str(p))
    if 0 < len(pkg_path_list):
        subprocess.run(["pip", "install", "--no-deps", *pkg_path_list])

    if 0 < len({dependencies}):
        args = ["pip", "install", *{dependencies}]
        subprocess.run(args)

    # this is base64 encoded source code
    tar_io = io.BytesIO(gzip.decompress(base64.b64decode("{pkg_encoded}")))
    with TemporaryDirectory() as temp_dir:
        with tarfile.open(fileobj=tar_io) as tar:
            for member in tar.getmembers():
                pkg_path = Path(temp_dir) / f"{{member.name}}"
                content_bytes = tar.extractfile(member).read()
                pkg_path.write_bytes(content_bytes)
                os.system("pip install --no-deps {{pkg_path}}".format(pkg_path=pkg_path))

    sys.path.append("/kaggle/working")
    os.environ.update({env_variables})
__bootstrap__()"""


def create_bootstrap_code(
    pkg_encoded: str,
    pkg_dataset: str,
    env_variables: Dict,
    dependencies: Iterable[str],
    enable_internet: bool = False,
) -> str:
    return BOOTSTRAP_TEMPLATE.format(
        pkg_encoded=pkg_encoded,
        pkg_dataset=pkg_dataset,
        env_variables=json.dumps(env_variables),
        dependencies=json.dumps(dependencies),
        encoding="utf8",
    )



import base64
import gzip
import io
import tarfile
from typing import IO, Callable, Dict, Iterable

from ..exception import NotSupportedKernelType
from .kernels import KERNEL_CREATEORS
from .package import Package
from .packaging_system import build_packages, get_dependencies

Builder = Callable[[IO[str], str, Dict], str]


def create_encoded_archive(pkgs: Iterable[Package]) -> str:
    tar_output = io.BytesIO()
    with tarfile.TarFile(fileobj=tar_output, mode="w") as tar:
        for pkg in pkgs:
            info = tarfile.TarInfo(name=pkg.name)
            info.size = len(pkg.content)
            tar.addfile(info, io.BytesIO(pkg.content))

    compressed = gzip.compress(tar_output.getvalue(), compresslevel=9)
    return base64.b64encode(compressed).decode("utf-8")


def get_builder(kernel_type: str) -> Builder:
    if kernel_type not in KERNEL_CREATEORS:
        raise NotSupportedKernelType(kernel_type)

    def _builder(
        kernel_body_io: IO[str],
        pkg_dataset: str,
        env_variables: Dict,
        enable_internet: bool,
        enable_constraint: bool,
    ) -> str:
        encoded_archive = create_encoded_archive(build_packages())
        dependencies = get_dependencies(enable_constraint)
        kernel_body = kernel_body_io.read()
        return KERNEL_CREATEORS[kernel_type](
            kernel_body,
            pkg_encoded=encoded_archive,
            pkg_dataset=pkg_dataset,
            env_variables=env_variables,
            dependencies=dependencies,
            enable_internet=enable_internet,
        )

    return _builder



from typing import Dict


def merge(source: Dict, destination: Dict) -> Dict:
    """
    Merge given dictionaries recursively.
    Reference: https://stackoverflow.com/questions/20656135/python-deep-merge-dictionary-data

    >>> a = { 'a' : { 'b' : { 'd' : 1, 'e' : 2 } } }
    >>> b = { 'a' : { 'b' : { 'f' : 3, 'g' : 4 } } }
    >>> merge(a, b)
    {'a': {'b': {'f': 3, 'g': 4, 'd': 1, 'e': 2}}}

    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            merge(value, node)
        else:
            if key not in destination:
                destination[key] = value

    return destination



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List

from pkg_resources.extern.packaging.specifiers import SpecifierSet
from poetry.factory import Factory
from poetry.io.null_io import NullIO
from poetry.masonry.builders import WheelBuilder
from poetry.utils.env import NullEnv

from .package import Package


def _get_pkg_name(require, enable_constraint: bool = False):
    if enable_constraint or require.is_vcs():
        return require.base_pep_508_name
    return require.name


def get_dependencies(enable_constraint: bool = False) -> List[str]:
    poetry = Factory().create_poetry(Path.cwd())
    dependencies = [
        _get_pkg_name(r, enable_constraint) for r in poetry.package.requires
    ]
    return dependencies


def build_packages() -> List[Package]:
    poetry = Factory().create_poetry(Path.cwd())
    env = NullEnv()
    io = NullIO()

    with TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        wheel_pkg_name = WheelBuilder.make_in(poetry, env, io, temp_dir)
        pkg_path = temp_dir / wheel_pkg_name
        pkg_bytes = pkg_path.read_bytes()
        pkgs = [Package(wheel_pkg_name, pkg_bytes)]

    return pkgs



from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional

import click
from kaggle import KaggleApi
from kaggle.models.kaggle_models_extended import KernelPushResponse
from kaggle.models.kernel_push_request import KernelPushRequest

from ..exception import KktSectionNotFound
from ..parser import DEFAULT_KKT_CONFIG, KktParser
from .kkt_command import kkt_command


class KernelType(Enum):
    script = "script"
    notebook = "notebook"

    def __str__(self):
        return self.value


def competition_prompt(api: KaggleApi) -> str:
    competition_query = click.prompt("competition", default="", show_default=False)
    competitions = api.competitions_list(search=competition_query)
    for i, c in enumerate(competitions):
        click.echo(f"{i} {c}")
    competition_index = click.prompt(
        ">", type=int, show_choices=False, prompt_suffix=" "
    )
    return str(competitions[competition_index])


def dataset_prompt() -> List:
    dataset_sources = []
    while click.confirm("Would you like to add dataset sources?"):
        dataset_source = click.prompt("dataset source")
        dataset_sources.append(dataset_source)
    return dataset_sources


def init_impl(api: KaggleApi) -> Dict:
    meta_data: Dict[str, Any] = {}
    meta_data["competition"] = competition_prompt(api)
    meta_data["slug"] = click.prompt("slug", type=str)
    meta_data["code_file"] = click.prompt("code_file", default="script.py")
    meta_data["kernel_type"] = str(
        click.prompt(
            "kernel_type",
            default=DEFAULT_KKT_CONFIG["meta_data"]["kernel_type"],
            type=KernelType,
        )
    )
    meta_data["is_private"] = click.confirm(
        "is_private", default=DEFAULT_KKT_CONFIG["meta_data"]["is_private"]
    )
    meta_data["enable_gpu"] = click.confirm(
        "enable_gpu", default=DEFAULT_KKT_CONFIG["meta_data"]["enable_gpu"]
    )
    meta_data["enable_internet"] = click.confirm(
        "enable_internet", default=DEFAULT_KKT_CONFIG["meta_data"]["enable_internet"]
    )
    meta_data["dataset_sources"] = dataset_prompt()
    meta_data["competition_sources"] = [meta_data["competition"]]
    meta_data["enable_constraint"] = click.confirm(
        "enable_constraint",
        default=DEFAULT_KKT_CONFIG["meta_data"]["enable_constraint"],
    )

    kkt_config: Dict[str, Any] = {
        "meta_data": meta_data,
    }
    kkt_config["enable_git_tag"] = click.confirm(
        "enable_git_tag", default=DEFAULT_KKT_CONFIG["enable_git_tag"]
    )
    return kkt_config


def confirm_initialize() -> bool:
    return click.confirm(
        "Kkt section is found in pyproject.yml. Do you want to continue?"
    )


def get_project_name(parser: KktParser) -> Optional[str]:
    return parser.read_all().get("tool", {}).get("poetry", {}).get("name")


def has_package_source(project_root_path: Path, project_name: str) -> bool:
    pkg_dir = project_root_path / project_name
    src_pkg_dir = project_root_path / "src" / project_name
    return pkg_dir.exists() or src_pkg_dir.exists()


def confirm_add_package_source(project_name: str):
    return click.confirm("Would you like to add {project_name} package direcctory?")


def add_package_source(project_root_path: Path, project_name: str) -> None:
    pkg_dir_path = project_root_path / "src" / project_name
    print(pkg_dir_path)
    pkg_dir_path.mkdir(parents=True, exist_ok=True)

    init_file_path = pkg_dir_path / "__init__.py"
    print(init_file_path)
    init_file_path.touch()


@kkt_command(init=True)
def init(
    api: KaggleApi, kkt: Dict, pyproject_path: Path, *args: List, **kwargs: Dict
) -> None:
    if kkt is not None and not confirm_initialize():
        return
    click.echo("Appending Kkt section into your pyproject.toml config.")

    kkt = init_impl(api)

    parser = KktParser(pyproject_path)
    parser.write(kkt)

    project_name = get_project_name(parser)
    print(project_name)
    if project_name is None:
        print("no name")
        return

    project_root_path = pyproject_path.parent
    if has_package_source(project_root_path, project_name):
        print("has")
        return

    if confirm_add_package_source(project_name):
        print("add")
        add_package_source(project_root_path, project_name)



from pathlib import Path
from typing import Any, Dict, List

import click
from kaggle import KaggleApi

from .. import kernel_proc
from ..exception import MetaDataNotFound
from .kkt_command import kkt_command
from ..fetch import PackageLocation, fetch_packages


def _get_package_locations(list_response: Dict[str, Any]) -> List[PackageLocation]:
    return [
        PackageLocation(item["url"], item["fileName"])
        for item in list_response["files"]
    ]


def _print_paths(paths: List[Path]) -> None:
    for p in paths:
        click.echo(f"save to:{str(p)}")


@kkt_command()
@click.argument("dst_dir")
def download(
    api: KaggleApi, kkt: Dict, pyproject_path: Path, quiet: bool = False, **kwargs
) -> None:
    if "meta_data" not in kkt:
        raise MetaDataNotFound()
    meta_data = kkt["meta_data"].value

    dst_dir = kwargs.get("dst_dir")
    if dst_dir is None:
        raise ValueError("DST_DIR is None")
    dst_dir_path = Path(dst_dir)

    kernel_slug = meta_data["slug"]
    status = kernel_proc.status(api, kernel_slug)
    if status["status"] != "complete":
        click.echo("Kernel has not been completed yet.")
        return

    list_response = kernel_proc.list_outputs(api, kernel_slug)
    pkg_locations = _get_package_locations(list_response)
    download_files = fetch_packages(pkg_locations, dst_dir_path, quiet)
    for p in download_files:
        click.echo(f"save to:{str(p)}")



from typing import Dict

from kaggle import KaggleApi


def get_username(api: KaggleApi) -> str:
    return api.config_values[api.CONFIG_NAME_USER]


def get_slug(api: KaggleApi, meta_data: Dict) -> str:
    user_name = get_username(api)
    slug = meta_data["slug"]
    return f"{user_name}/{slug}"


def get_dataset_slug(api: KaggleApi, meta_data: Dict) -> str:
    slug = get_slug(api, meta_data)
    return f"{slug}-requirements"



import re
from pathlib import Path
from typing import Dict

from git import Repo as GitRepo
from tomlkit import dumps as dumps_as_toml

from .exception import FoundUncommitedFiles


def _create_tag(version: int) -> str:
    return "kernel_version_{}".format(version)


def _create_message(meta_data: Dict, env_variables: Dict) -> str:
    meta_data_dump = dumps_as_toml(meta_data)
    env_variables_dump = dumps_as_toml(env_variables)
    return f"[tool.kkt.meta_data]\n{meta_data_dump}\n\n[tool.kkt.environment_variables]\n{env_variables_dump}"


class Repo:
    def __init__(self, path: Path) -> None:
        self.path = path

    @property
    def git_repo(self) -> GitRepo:
        return GitRepo(self.path)

    def _check_uncommited_files(self) -> None:
        untracked_files = self.git_repo.untracked_files
        modified_files = [_.a_path for _ in self.git_repo.index.diff(None)]
        staged_files = [_.a_path for _ in self.git_repo.index.diff("HEAD")]
        uncommitted_files = [*untracked_files, *modified_files, *staged_files]
        if 0 < len(uncommitted_files):
            raise FoundUncommitedFiles(uncommitted_files)

    def validate(self) -> None:
        self._check_uncommited_files()

    def attach_version_tag(
        self, version: int, meta_data: Dict, env_variables: Dict
    ) -> None:
        tag = _create_tag(version)
        message = _create_message(meta_data, env_variables)
        self.git_repo.create_tag(tag, message=message)



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Any
from functools import singledispatch

import click
from kaggle import KaggleApi
from kaggle.models.dataset_new_request import DatasetNewRequest
from kaggle.models.dataset_new_version_request import DatasetNewVersionRequest
from kaggle.models.kaggle_models_extended import (
    DatasetNewResponse,
    DatasetNewVersionResponse,
    KernelPushResponse,
)
from kaggle.models.kernel_push_request import KernelPushRequest

from .resource import get_slug, get_username


@dataclass
class KernelPushParams:
    id_no: Optional[str]
    slug: str
    new_title: Optional[str]
    kernel_type: str
    is_private: Optional[bool]
    enable_gpu: Optional[bool]
    enable_internet: Optional[bool]
    dataset_data_sources: List[str]
    competition_data_sources: List[str]
    kernel_data_sources: List[str]
    category_ids: List[str]

    @classmethod
    def of(cls, api, meta_data: Dict) -> "KernelPushParams":
        id_no = meta_data.get("id_no")
        slug = get_slug(api, meta_data)
        new_title = api.get_or_default(meta_data, "slug", None)
        kernel_type = meta_data.get("kernel_type", "script")
        is_private = api.get_or_default(meta_data, "is_private", None)
        enable_gpu = api.get_or_default(meta_data, "enable_gpu", None)
        enable_internet = api.get_or_default(meta_data, "enable_internet", None)

        dataset_data_sources = api.get_or_default(meta_data, "dataset_sources", [])
        for source in dataset_data_sources:
            api.validate_dataset_string(source)

        competition_data_sources = api.get_or_default(
            meta_data, "competition_sources", []
        )

        kernel_data_sources = api.get_or_default(meta_data, "kernel_sources", [])
        for source in kernel_data_sources:
            api.validate_kernel_string(source)

        category_ids = api.get_or_default(meta_data, "keywords", [])

        return cls(
            id_no=id_no,
            slug=slug,
            new_title=new_title,
            kernel_type=kernel_type,
            is_private=is_private,
            enable_gpu=enable_gpu,
            enable_internet=enable_internet,
            dataset_data_sources=dataset_data_sources,
            competition_data_sources=competition_data_sources,
            kernel_data_sources=kernel_data_sources,
            category_ids=category_ids,
        )


def push(
    api: KaggleApi, params: KernelPushParams, script_body: str
) -> KernelPushResponse:
    """ read the metadata file and kernel files from a notebook, validate
        both, and use Kernel API to push to Kaggle if all is valid.
         Parameters
        ==========
        folder: the path of the folder
    """
    language = "python"
    kernel_push_request = KernelPushRequest(
        id=params.id_no,
        slug=params.slug,
        new_title=params.new_title,
        text=script_body,
        language=language,
        kernel_type=params.kernel_type,
        is_private=params.is_private,
        enable_gpu=params.enable_gpu,
        enable_internet=params.enable_internet,
        dataset_data_sources=params.dataset_data_sources,
        competition_data_sources=params.competition_data_sources,
        kernel_data_sources=params.kernel_data_sources,
        category_ids=params.category_ids,
    )

    result = KernelPushResponse(
        api.process_response(
            api.kernel_push_with_http_info(kernel_push_request=kernel_push_request)
        )
    )
    return result


def status(api: KaggleApi, kernel_slug: str):
    user_name = api.config_values[api.CONFIG_NAME_USER]
    return api.kernel_status(user_name, kernel_slug)


def list_outputs(api: KaggleApi, kernel_slug: str):
    user_name = api.config_values[api.CONFIG_NAME_USER]
    return api.process_response(
        api.kernel_output_with_http_info(user_name, kernel_slug)
    )


def create_dataset(
    api: KaggleApi,
    dataset_slug: str,
    license_name: str,
    target_dir: Path,
    quiet: bool = False,
):
    if len(dataset_slug) < 6 or len(dataset_slug) > 50:
        raise ValueError("The dataset slug must be between 6 and 50 characters")

    owner_slug = get_username(api)
    request = DatasetNewRequest(
        title=dataset_slug,
        slug=dataset_slug,
        owner_slug=owner_slug,
        license_name=license_name,
        subtitle=None,
        description=None,
        files=[],
        is_private=True,
        convert_to_csv=False,
        category_ids=[],
    )
    api.upload_files(request, None, target_dir, quiet)

    result = DatasetNewResponse(
        api.process_response(api.datasets_create_new_with_http_info(request))
    )
    return result


def update_dataset(
    api: KaggleApi,
    dataset_slug: str,
    target_dir: Path,
    quiet=False,
    delete_old_versions=True,
):
    owner_slug = get_username(api)
    request = DatasetNewVersionRequest(
        version_notes="test",
        subtitle=None,
        description=None,
        files=[],
        convert_to_csv=False,
        category_ids=[],
        delete_old_versions=delete_old_versions,
    )
    api.upload_files(request, None, target_dir, quiet)

    result = DatasetNewVersionResponse(
        api.process_response(
            api.datasets_create_version_with_http_info(
                owner_slug, dataset_slug, request
            )
        )
    )
    return result


@singledispatch
def print_response(result: Any) -> None:
    raise NotImplementedError(f"print_response not support {type(result)}")


@print_response.register
def _print_KernelPushResponse(result: KernelPushResponse) -> None:
    if result.error is not None:
        click.echo("error: {}".format(result.error))
    else:
        click.echo("ref: {}".format(result.ref))
        click.echo("url: {}".format(result.url))
        click.echo("version: {}".format(result.versionNumber))


@print_response.register
def _print_DatasetNewResponse(result: DatasetNewResponse) -> None:
    if result.status == "error":
        click.echo("error: {}".format(result.error))
    else:
        click.echo("ref: {}".format(result.ref))
        click.echo("url: {}".format(result.url))


@print_response.register
def _print_DatasetNewVersionResponse(result: DatasetNewVersionResponse) -> None:
    if result.status == "error":
        click.echo("error: {}".format(result.error))
    else:
        click.echo("ref: {}".format(result.ref))
        click.echo("url: {}".format(result.url))



from typing import Dict, Iterable

from .bootstrap import create_bootstrap_code


SCRIPT_TEMPLATE: str = """{bootstrap_code}

{script_body}
"""


def create_script_kernel(
    script_body: str,
    pkg_encoded: str,
    pkg_dataset: str,
    env_variables: Dict,
    dependencies: Iterable[str],
    enable_internet: bool = False,
) -> str:
    bootstrap_code = create_bootstrap_code(
        pkg_encoded=pkg_encoded,
        pkg_dataset=pkg_dataset,
        env_variables=env_variables,
        dependencies=dependencies,
        enable_internet=enable_internet,
    )
    return SCRIPT_TEMPLATE.format(
        bootstrap_code=bootstrap_code, script_body=script_body, encoding="utf8"
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kkt', 'kkt.builders', 'kkt.builders.kernels', 'kkt.commands']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0',
 'gitpython>=3.1,<4.0',
 'kaggle>=1.5,<2.0',
 'lockfile>=0.12.2,<0.13.0',
 'poetry==1.0.0',
 'tomlkit>=0.5.8,<0.6.0']

entry_points = \
{'console_scripts': ['kkt = kkt.cli:main']}

setup_kwargs = {
    'name': 'kkt',
    'version': '0.3.4',
    'description': 'A tool for kaggle kernel',
    'long_description': '# kkt\nkkt is a tool for kaggle kernel management.\n\n[![Actions Status](https://github.com/ar90n/kkt/workflows/Python%20package/badge.svg)](https://github.com/ar90n/kkt/actions)\n[![PyPI](https://img.shields.io/pypi/v/kkt.svg)](https://pypi.python.org/pypi/kkt)\n[![PythonVersions](https://img.shields.io/pypi/pyversions/kkt.svg)](https://pypi.python.org/pypi/kkt)\n\n## Feature\n* Show the status of the latest version\n* Push your script or notebook to the Kaggle Kernels\n* Pack and emmbedded your library codes into the generated bootstrap codes\n* Create a dataset containing your dependent packages\n* Add bootstrap codes into the head of your script or notebook automatically\n* Add git tags whose name is corresponding kernel version\n* Set environment variable for your kernels\n\n## Installation\nFor now, kkt is designed to be used with poetry. So kkt can be installed by the following.\n\n```bash\n$ poetry add kkt --dev\n```\n\n## Usage\n\n### Set username and token of kaggle-api\nPlease setup your kaggle-api credentials as following this [article](https://github.com/Kaggle/kaggle-api#api-credentials)\n\n### Setup kkt in your project\nSetup this project for [digit-recognizer competition](https://www.kaggle.com/c/digit-recognizer).\nIn this configuration, we use script.py. If you want to use notebook, kkt also support it.\n\n```bash\n$ poetry run kkt init\nAppending Kkt section into your pyproject.toml config.\ncompetition: digit\n0 digit-recognizer\n> 0\nslug: kkt-example\ncode_file [script.py]: script.py\nkernel_type [script]: script\nis_private [Y/n]: n\nenable_gpu [y/N]: n\nenable_internet [y/N]: y\nWould you like to add dataset sources? [y/N]: n\nenable_git_tag [y/N]: n\n```\n\n### Create kkt_example package and its driver code.\nkkt_example provides random choice solver for digit-recognizer competition.\n```bash\n$ tree\n.\n├── kkt_example\n│\xa0\xa0 └── __init__.py\n├── poetry.lock\n├── pyproject.toml\n└── script.py\n\n1 directory, 4 files\n```\n\n__init__.py\n```python\nfrom pathlib import Path\nimport random\n\nimport pandas as pd\n\ndef choice():\n    return random.randint(0, 9)\n\ndef load_sample_submission():\n    path = Path("..") / "input" / "digit-recognizer" / "sample_submission.csv"\n    return pd.read_csv(path,  index_col="ImageId")\n```\n\nscript.py\n```python\nimport kkt_example\n\nsubmission = kkt_example.load_sample_submission()\nfor _, row in submission.iterrows():\n    row["Label"] = kkt_example.choice()\n\n    submission.to_csv("submission.csv")\n```\n\npyproject.toml\n```toml\n[tool.poetry]\nname = "kkt-example"\nversion = "0.1.0"\ndescription = ""\nauthors = ["Masahiro Wada <argon.argon.argon@gmail.com>"]\n\n[tool.poetry.dependencies]\npython = "^3.7"\npandas = "^1.0.0"\n\n[tool.poetry.dev-dependencies]\nkkt = "^0.3.1"\n\n\n[tool.kkt]\nenable_git_tag = false\n\n[tool.kkt.meta_data]\ncode_file = "script.py"\ncompetition = "digit-recognizer"\ncompetition_sources = ["digit-recognizer"]\ndataset_sources = []\nenable_gpu = false\nenable_internet = true\nis_private = false\nkernel_type = "script"\nslug = "kkt-example"\n\n[build-system]\nrequires = ["poetry>=0.12"]\nbuild-backend = "poetry.masonry.api"\n```\n\nIf you want run script.py in local environmet, please run the following.\n\n```bash\n$ poetry run python script.py\n$ head submission.csv\nImageId,Label\n1,1\n2,1\n3,0\n4,2\n5,4\n6,8\n7,5\n8,3\n9,2\n```\n\n### Create a dataset containing dependent packages if need\nIn this example, there aren\'t extra required packages. So `kkt install` displays the following. And this step is not mandatory.\n\n```bash\n$ poetry run kkt install\nref: /ar90ngas/kkt-example-install\nurl: https://www.kaggle.com/ar90ngas/kkt-example-install\nversion: 1\nPushing install kernel successed.\nWait for install kernel completion...\nWait for install kernel completion...\nWait for install kernel completion...\nExtra required packages are nothing.\n```\n\nBut in the little complicated project such as mnist_efficientnet example, `kkt install` displays the following. This means that a new dataset whose slug is `ar90ngas/mnist-efficientnet-requirements` is created. And it contains an extra package which is required by this example. And this package will be installed automatically in the bootstrap code.\n\n```bash\n$ poetry run kkt install\nref: /ar90ngas/mnist-efficientnet-install\nurl: https://www.kaggle.com/ar90ngas/mnist-efficientnet-install\nversion: 1\nPushing install kernel successed.\nWait for install kernel completion...\nWait for install kernel completion...\nWait for install kernel completion...\nOutput file downloaded to /tmp/tmpq6m9iq9p/timm-0.1.30-py3-none-any.whl\nStarting upload for file timm-0.1.30-py3-none-any.whl\n100%|█████████████████████████████████████████████████████████| 203k/203k [00:03<00:00, 53.7kB/s]\nUpload successful: timm-0.1.30-py3-none-any.whl (203KB)\nref: ar90ngas/mnist-efficientnet-requirements\nurl: https://www.kaggle.com/ar90ngas/mnist-efficientnet-requirements\n```\n\n### Push notebook to Kaggle Kernels\n```bash\n$ poetry run kkt push\nref: /ar90ngas/kkt-example\nurl: https://www.kaggle.com/ar90ngas/kkt-example\nversion: 1\n```\nPlease visit [the result](https://www.kaggle.com/ar90ngas/kkt-example).\n\n### Show the status\n```bash\n$ poetry run kkt status\nstatus: complete\n```\n\n## Configuration\nPlease see [examples](https://github.com/ar90n/kkt/tree/master/examples)\n\n\n## License\nThis software is released under the Apache License, see [LICENSE](LICENSE).\n',
    'author': 'Masahiro Wada',
    'author_email': 'argon.argon.argon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ar90n/kkt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import re
from pathlib import Path
from typing import Any, Dict, List, Union

from tomlkit.exceptions import NonExistentKey
from tomlkit.toml_file import TOMLFile

from .exception import (
    InvalidTarget,
    KktSectionNotFound,
    MandatoryKeyNotFound,
    MetaDataNotFound,
)
from .util import merge

MANDATORY_KEYS: List = [("meta_data", [("slug", []), ("code_file", [])])]

DEFAULT_KKT_CONFIG: Dict = {
    "meta_data": {
        "kernel_type": "script",
        "is_private": True,
        "enable_gpu": False,
        "enable_internet": False,
        "dataset_sources": [],
        "kernel_data_sources": [],
        "competition_sources": [],
        "keywords": [],
        "enable_constraint": False,
    },
    "enable_git_tag": False,
}

META_DATA_KEYS = set(DEFAULT_KKT_CONFIG["meta_data"].keys())
META_DATA_KEYS.add("slug")
META_DATA_KEYS.add("code_file")
META_DATA_KEYS.add("competition")


def _validate_keys(obj: Dict[str, Any], mandatory_keys: List, path: str = "") -> None:
    """
    Raise MandatoryKeyNotFound if given obj doesn't contain all elements of mandatory_keys.

    >>> _validate_keys({"meta_data": {"slug": "slug", "code_file": "code_file"}}, MANDATORY_KEYS)

    # Comment out because pytest integration joesn't work
    #>>> _validate_keys({"meta_data": {"slug": "slug"}}, MANDATORY_KEYS)
    #MandatoryKeyNotFound('.meta_data.slug.code_file',)
    """

    for cur, children in mandatory_keys:
        path = "{}.{}".format(path, cur)
        if cur not in obj:
            raise MandatoryKeyNotFound(path)
        _validate_keys(obj[cur], children, path=path)


def _get_meta_data_keys(key: str) -> List[str]:
    """
    Return a list of keys to fetch partial meta_data from meta_data root.

    >>> _get_meta_data_keys(".")
    ['meta_data']
    >>> _get_meta_data_keys(".a.b.c")
    ['meta_data', 'a', 'b', 'c']
    """

    if re.match(r"^(\.|(\.[^.]+)+)$", key) is None:
        raise InvalidTarget(key)

    # key == "." means the root of meta_data
    return ["meta_data"] if key == "." else f"meta_data{key}".split(".")


def _trim_redundant_fields(meta_data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Trim field whose key are not in META_DATA_KEYS.

    >>> _trim_redundant_fields({"slug": "slug", "redundant": "redundant"})
    {'slug': 'slug'}
    """

    return {k: v for k, v in meta_data.items() if k in META_DATA_KEYS}


def _compose_meta_data(kkt: Dict[str, Any], key: str) -> Dict[str, Any]:
    """
    Compose a meta_data dict by recurcive sampling and merging them.

    >>> kkt = {                                \
        "meta_data": {                         \
            "slug": "slug",                    \
            "nest1": {                         \
                "code_file": "script.py",      \
                "nest2": {                     \
                    "kernel_type" : "script"   \
                }                              \
            }                                  \
        }                                      \
    }
    >>> _compose_meta_data(kkt, ".nest1.nest2")
    {'slug': 'slug', 'code_file': 'script.py', 'kernel_type': 'script'}
    """

    def _aux(meta_data: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
        if len(keys) == 0:
            return meta_data

        try:
            meta_data = meta_data[keys[0]]
        except KeyError:
            raise MetaDataNotFound(key)

        return {**meta_data, **_aux(meta_data, keys[1:])}

    keys = _get_meta_data_keys(key)
    return _trim_redundant_fields(_aux(kkt, keys))


class PyprojectParser(TOMLFile):
    def __init__(self, path: Union[str, Path]) -> None:
        super().__init__(str(path))

        self._path = Path(path)

    @property
    def path(self) -> Path:
        return self._path


class KktParser(PyprojectParser):
    def __init__(self, path: Union[str, Path]) -> None:
        super().__init__(path)

    def read_all(self) -> Dict[str, Any]:
        return super().read()

    def read(self, key: str = ".") -> Dict[str, Any]:
        pyproj = super().read()
        try:
            kkt = pyproj["tool"]["kkt"]
        except (NonExistentKey, KeyError):
            raise KktSectionNotFound()

        # following code didn't work.
        # kkt["meta_data"] =  _compose_meta_data(kkt, key)
        tmp = _compose_meta_data(kkt, key)
        del kkt["meta_data"]
        kkt.add("meta_data", tmp)

        kkt = merge(DEFAULT_KKT_CONFIG, kkt)
        _validate_keys(kkt, MANDATORY_KEYS)

        return kkt

    def write(self, kkt_config: Dict[str, Any]) -> None:
        pyproj = super().read()
        pyproj.setdefault("tool", {})["kkt"] = kkt_config

        super().write(pyproj)



class FoundUncommitedFiles(EnvironmentError):
    pass


class KktSectionNotFound(EnvironmentError):
    pass


class MandatoryKeyNotFound(ValueError):
    pass


class MetaDataNotFound(EnvironmentError):
    pass


class NotSupportedKernelType(ValueError):
    pass


class InvalidTarget(ValueError):
    pass


class InstallKernelError(RuntimeError):
    pass
